use serde::Serialize;
use serde_json::{Value, json};

pub mod gpu {
    use serde::Serialize;
    use serde_json::{Value, json};
    use wgpu;
    use winit::{
        application::ApplicationHandler,
        event::WindowEvent,
        event_loop::{ActiveEventLoop, EventLoop},
        window::WindowId,
    };

    #[derive(Debug, Serialize)]
    pub struct DriverInfo {
        version: String,
        name: String,
    }

    #[derive(Debug, Serialize)]
    pub struct AdapterInfo {
        vendor: String,
        model: String,
        driver: DriverInfo,
        vram: String,
        display: DisplayInfo,
    }

    #[derive(Debug, Serialize)]
    pub struct GPUInfo {
        supported_backends: Vec<String>,
        gpus: Vec<AdapterInfo>,
    }

    #[derive(Debug, Serialize, Default)]
    pub struct DisplayInfo {
        resolution: String,
        refresh_rate: String,
    }

    impl ApplicationHandler for DisplayInfo {
        fn resumed(&mut self, event_loop: &ActiveEventLoop) {
            if let Some(monitor) = event_loop.primary_monitor() {
                let size = monitor.size();
                let refresh_rate = monitor.refresh_rate_millihertz();

                self.resolution = format!("{}x{}", size.width, size.height);
                self.refresh_rate = if let Some(refresh) = refresh_rate {
                    format!("{} hz", refresh / 1000)
                } else {
                    String::from("Unknown")
                }
            } else {
                self.resolution = String::from("Unknown");
                self.refresh_rate = String::from("Unknown");
            }

            event_loop.exit();
        }

        fn window_event(
            &mut self,
            _event_loop: &ActiveEventLoop,
            _window_id: WindowId,
            _event: WindowEvent,
        ) {
        }
    }

    mod vram {
        use ash::vk::{API_VERSION_1_2, ApplicationInfo, InstanceCreateInfo, MemoryHeapFlags};
        use humansize::{DECIMAL, make_format};

        #[cfg(not(target_os = "macos"))]
        pub fn get_metal() -> u64 {
            0
        }

        #[cfg(target_os = "macos")]
        pub fn get_metal() -> u64 {
            use metal::Device as MetalDevice;

            let device = MetalDevice::system_default().expect("No Metal-compatible GPU found");
            device.recommended_max_working_set_size()
        }

        pub fn get_vulkan(device_id: u32) -> u64 {
            let entry = unsafe { ash::Entry::load().unwrap() };
            let app_info = ApplicationInfo {
                p_application_name: std::ptr::null(),
                application_version: 0,
                p_engine_name: std::ptr::null(),
                engine_version: 0,
                api_version: API_VERSION_1_2,
                ..Default::default()
            };

            let create_info = InstanceCreateInfo {
                p_application_info: &app_info,
                ..Default::default()
            };

            let instance = unsafe { entry.create_instance(&create_info, None).unwrap() };

            let physical_devices = unsafe { instance.enumerate_physical_devices().unwrap() };
            let mut total_vram = 0;

            for device in physical_devices {
                let memory_properties =
                    unsafe { instance.get_physical_device_memory_properties(device) };
                let device_properties = unsafe { instance.get_physical_device_properties(device) };

                if device_id != device_properties.device_id {
                    continue;
                }

                for heap in memory_properties.memory_heaps {
                    if heap.flags.contains(MemoryHeapFlags::DEVICE_LOCAL) {
                        total_vram += heap.size;
                    }
                }

                break;
            }

            total_vram
        }

        pub fn get(vendor: u32, device_id: u32) -> String {
            let formatter = make_format(DECIMAL);

            match vendor {
                0x10DE | 0x1002 | 0x8086 | 0x5143 => formatter(get_vulkan(device_id)),
                0x1010 => formatter(get_metal()),
                _ => String::from("Unknown"),
            }
        }
    }

    fn vendor_name(vendor: u32) -> &'static str {
        match vendor {
            0x10DE => "NVIDIA",
            0x1002 => "AMD(Advanced Micro Devices), Inc.",
            0x8086 => "Intel(integrated electronics)",
            0x13B5 => "ARM(Advanced RISC Machines)",
            0x5143 => "Qualcomm(Quality Communications)",
            0x1010 => "Apple Inc.",
            _ => "Unknown",
        }
    }

    pub fn get_struct() -> GPUInfo {
        let mut gpu_data: Vec<AdapterInfo> = Vec::new();
        let mut backends: Vec<String> = Vec::new();
        let instance_descriptor = wgpu::InstanceDescriptor {
            backends: wgpu::Backends::all(),
            flags: wgpu::InstanceFlags::empty(),
            backend_options: Default::default(),
        };
        let instance = wgpu::Instance::new(&instance_descriptor);
        let event_loop = EventLoop::new().unwrap();
        let mut app = DisplayInfo::default();

        event_loop.run_app(&mut app).unwrap();

        for adapter in instance.enumerate_adapters(wgpu::Backends::all()) {
            let info = adapter.get_info();

            if !backends.contains(&info.backend.to_string()) {
                backends.push(info.backend.to_string());
            }
            if info.driver.is_empty() || info.driver_info.is_empty() {
                continue;
            }

            gpu_data.push(AdapterInfo {
                vendor: vendor_name(info.vendor).to_string(),
                model: info.name,
                driver: DriverInfo {
                    version: info.driver_info,
                    name: info.driver,
                },
                vram: vram::get(info.vendor, info.device),
                display: DisplayInfo {
                    resolution: app.resolution.to_string(),
                    refresh_rate: app.refresh_rate.to_string(),
                },
            });
        }

        GPUInfo {
            supported_backends: backends,
            gpus: gpu_data,
        }
    }

    pub fn get_json() -> Value {
        json!(get_struct())
    }
}

pub mod cpu {
    use serde::Serialize;
    use serde_json::{Value, json};
    use std::collections::HashMap;
    use sysinfo::{CpuRefreshKind, RefreshKind, System};

    #[derive(Debug, Serialize)]
    pub struct ProcessorInfo {
        vendor: String,
        brand: String,
        total_system_cores: String,
        threads: u16,
        architecture: String,
        byte_order: String,
    }

    #[derive(Debug, Serialize)]
    pub struct CPUInfo {
        cpus: Vec<ProcessorInfo>,
    }

    pub fn get_struct() -> CPUInfo {
        let system = System::new_with_specifics(
            RefreshKind::nothing().with_cpu(CpuRefreshKind::everything()),
        );

        let mut processors = Vec::new();
        let mut processor_count: HashMap<String, ProcessorInfo> = HashMap::new();

        for cpu in system.cpus() {
            let brand = cpu.brand().trim();
            let total_cores = System::physical_core_count();

            if processor_count.contains_key(brand) {
                processor_count.get_mut(brand).unwrap().threads += 1;
                continue;
            }
            processor_count.insert(
                brand.to_string(),
                ProcessorInfo {
                    vendor: cpu.vendor_id().trim().to_string(),
                    brand: brand.to_string(),
                    total_system_cores: if let Some(cores) = total_cores {
                        cores.to_string()
                    } else {
                        String::from("Unknown")
                    },
                    threads: 1,
                    architecture: System::cpu_arch().trim().to_string(),
                    byte_order: if cfg!(target_endian = "little") {
                        String::from("little-endian")
                    } else {
                        String::from("big-endian")
                    },
                },
            );
        }

        for (_brand, info) in processor_count {
            processors.push(info);
        }

        CPUInfo { cpus: processors }
    }
    pub fn get_json() -> Value {
        json!(get_struct())
    }
}

pub mod memory {
    use humansize::{DECIMAL, make_format};
    use serde::Serialize;
    use serde_json::{Value, json};
    use sysinfo::System;

    #[derive(Debug, Serialize)]
    pub struct PhysicalInfo {
        total: String,
        used: String,
        free: String,
        available: String,
    }

    #[derive(Debug, Serialize)]
    pub struct VirtualInfo {
        total: String,
        used: String,
        free: String,
    }

    #[derive(Debug, Serialize)]
    pub struct MemoryInfo {
        physical: PhysicalInfo,
        virtual_swap: VirtualInfo,
    }

    pub fn get_struct() -> MemoryInfo {
        let system = System::new_all();
        let formatter = make_format(DECIMAL);

        MemoryInfo {
            physical: PhysicalInfo {
                total: formatter(system.total_memory()),
                used: formatter(system.used_memory()),
                free: formatter(system.free_memory()),
                available: formatter(system.available_memory()),
            },
            virtual_swap: VirtualInfo {
                total: formatter(system.total_swap()),
                used: formatter(system.used_swap()),
                free: formatter(system.free_swap()),
            },
        }
    }

    pub fn get_json() -> Value {
        json!(get_struct())
    }
}

pub mod os {
    use detect_desktop_environment::DesktopEnvironment;
    use serde::Serialize;
    use serde_json::{Value, json};
    use sysinfo::System;

    #[derive(Debug, Serialize)]
    pub struct OSInfo {
        name: String,
        edition: String,
        version: String,
        architecture: String,
        kernel: String,
        de: String,
    }

    pub fn get_struct() -> OSInfo {
        OSInfo {
            name: System::name().unwrap_or(String::from("Unknown")),
            edition: System::long_os_version().unwrap_or(String::from("Unknown")),
            version: System::os_version().unwrap_or(String::from("Unknown")),
            architecture: if cfg!(target_pointer_width = "64") {
                String::from("64 bit")
            } else if cfg!(target_pointer_width = "32") {
                String::from("32 bit")
            } else {
                String::from("16 bit")
            },
            kernel: System::kernel_long_version(),
            de: match DesktopEnvironment::detect() {
                Some(de) => format!("{de:?}"),
                None => String::from(""),
            },
        }
    }

    pub fn get_json() -> Value {
        json!(get_struct())
    }
}

pub mod storage {
    use humansize::{DECIMAL, make_format};
    use serde::Serialize;
    use serde_json::{Value, json};
    use sysinfo::Disks;

    #[derive(Debug, Serialize)]
    pub struct StorageInfo {
        name: String,
        mount_point: String,
        kind: String,
        space_left: String,
        total: String,
    }

    pub fn get_list() -> Vec<StorageInfo> {
        let disks = Disks::new_with_refreshed_list();
        let mut drive_data = Vec::new();
        let formatter = make_format(DECIMAL);

        for disk in disks.list() {
            drive_data.push(StorageInfo {
                name: format!("{:?}", disk.name()).trim_matches('\"').to_string(),
                mount_point: format!("{:?}", disk.mount_point())
                    .replace("\\", "")
                    .trim_matches('\"')
                    .to_string(),
                kind: format!("{:?}", disk.kind()),
                space_left: formatter(disk.available_space()),
                total: formatter(disk.total_space()),
            });
        }

        drive_data
    }

    pub fn get_json() -> Value {
        json!(get_list())
    }
}

pub mod network {
    use serde::Serialize;
    use serde_json::{Value, json};
    use sysinfo::Networks;

    #[derive(Debug, Serialize)]
    pub struct NetworkInfo {
        adapters: Vec<String>,
    }

    pub fn get_struct() -> NetworkInfo {
        let networks = Networks::new_with_refreshed_list();
        let mut adapters = Vec::new();

        for (interface_name, _network_data) in &networks {
            adapters.push(interface_name.to_string());
        }

        NetworkInfo { adapters }
    }

    pub fn get_json() -> Value {
        json!(get_struct())
    }
}

pub mod external {
    use serde::Serialize;
    use serde_json::{Value, json};
    use std::{cmp::Ordering::Equal, collections::HashMap};
    use sysinfo::{System, Users};

    #[derive(Debug, Serialize)]
    pub struct SoftwareInfo {
        name: String,
        count: usize,
    }

    #[derive(Debug, Serialize)]
    pub struct ExternalInfo {
        softwares: Vec<SoftwareInfo>,
    }

    pub fn get_struct() -> ExternalInfo {
        let mut system = System::new_all();
        system.refresh_all();

        let users = &Users::new_with_refreshed_list();
        let mut grouped_processes: HashMap<String, Vec<i32>> = HashMap::new();

        for process in system.processes().values() {
            let name = process.name().to_str().unwrap().to_string();
            let user_id = process.user_id();
            let is_user_owned = user_id.is_some_and(|uid| {
                users
                    .list()
                    .iter()
                    .any(|u| u.id() == uid && u.name() != "root")
            });

            if is_user_owned && !name.trim().is_empty() && !name.starts_with('[') {
                grouped_processes
                    .entry(name)
                    .or_default()
                    .push(process.pid().as_u32() as i32);
            }
        }

        let mut softwares = Vec::new();
        let mut grouped_vec: Vec<_> = grouped_processes.into_iter().collect();
        grouped_vec.sort_by(|a, b| match b.1.len().cmp(&a.1.len()) {
            Equal => a.0.to_lowercase().cmp(&b.0.to_lowercase()),
            other => other,
        });

        for (name, pids) in grouped_vec {
            softwares.push(SoftwareInfo {
                name,
                count: pids.len(),
            });
        }

        ExternalInfo { softwares }
    }

    pub fn get_json() -> Value {
        json!(get_struct())
    }
}

pub mod devices {
    use serde::Serialize;
    use serde_json::{Value, json};
    extern crate hidapi;
    use hidapi::HidApi;
    use std::collections::HashMap;

    #[derive(Debug, Serialize)]
    pub struct DeviceInfo {
        manufacturer: String,
        products: Vec<String>,
    }

    #[derive(Debug, Serialize)]
    pub struct DevicesInfo {
        devices: Vec<DeviceInfo>,
    }

    pub fn get_struct() -> DevicesInfo {
        let api = HidApi::new().unwrap();
        let mut grouped_devices: HashMap<String, Vec<String>> = HashMap::new();

        for device in api.device_list() {
            let manufacturer = device
                .manufacturer_string()
                .unwrap_or("Unknown")
                .to_string()
                .to_lowercase();
            let product = device.product_string().unwrap_or("Unknown").to_string();

            if !manufacturer.trim().is_empty() && !product.trim().is_empty() {
                let entry = grouped_devices.entry(manufacturer).or_default();

                if !entry.contains(&product) {
                    entry.push(product);
                }
            }
        }

        let mut devices = Vec::new();
        let mut grouped_vec: Vec<_> = grouped_devices.into_iter().collect();
        grouped_vec.sort_by(|a, b| a.0.to_lowercase().cmp(&b.0.to_lowercase()));

        for (manufacturer, products) in grouped_vec {
            devices.push(DeviceInfo {
                manufacturer,
                products,
            });
        }

        DevicesInfo { devices }
    }

    pub fn get_json() -> Value {
        json!(get_struct())
    }
}

pub mod meta {
    use serde::Serialize;
    use serde_json::{Value, from_str, json};
    use std::{collections::HashMap, fs};
    use sys_locale::get_locale;

    include!(concat!(env!("OUT_DIR"), "/built.rs"));

    #[derive(Debug, Serialize)]
    pub struct MetaInfo {
        version: String,
        commit_hash: String,
        commit_short_hash: String,
        locale: String,
        language: String,
        territory: String,
    }

    pub fn get_struct() -> MetaInfo {
        let default = String::from("Unknown");
        let locale = get_locale().unwrap_or_else(|| String::from("en-US"));
        let split_locale: Vec<&str> = locale.split("-").collect();

        let language_json = fs::read_to_string("subcrates/telemetry/src/languages.json").unwrap();
        let parsed_languages = from_str::<HashMap<String, String>>(&language_json).unwrap();
        let language_map = parsed_languages.get(split_locale[0]).unwrap_or(&default);

        let territory_json =
            fs::read_to_string("subcrates/telemetry/src/territories.json").unwrap();
        let parsed_territories = from_str::<HashMap<String, String>>(&territory_json).unwrap();
        let territory_map = parsed_territories.get(split_locale[1]).unwrap_or(&default);

        MetaInfo {
            version: env!("CARGO_PKG_VERSION").to_string(),
            commit_hash: GIT_COMMIT_HASH.to_string(),
            commit_short_hash: GIT_COMMIT_SHORT_HASH.to_string(),
            locale,
            language: language_map.to_string(),
            territory: territory_map.to_string(),
        }
    }

    pub fn get_json() -> Value {
        json!(get_struct())
    }
}

pub mod uptime {
    use chrono::{DateTime, Utc};
    use serde::Serialize;
    use serde_json::{Value, json};
    use std::time::Duration as StdDuration;
    use sysinfo::System;

    #[derive(Debug, Serialize)]
    pub struct DateInfo {
        zone: String,
        date: String,
        time: String,
    }

    #[derive(Debug, Serialize)]
    pub struct UptimeInfo {
        boot: DateInfo,
        now: DateInfo,
        relative: String,
    }

    pub fn format_duration(duration: StdDuration) -> String {
        let mut parts = Vec::new();
        let mut remaining = duration.as_secs();
        let units = [
            (" months", 30 * 24 * 60 * 60),
            (" weeks", 7 * 24 * 60 * 60),
            (" days", 24 * 60 * 60),
            (" hours", 60 * 60),
            (" minutes", 60),
            (" seconds", 1),
        ];

        for &(label, unit_secs) in &units {
            match remaining / unit_secs {
                0 => continue,
                value => {
                    parts.push(format!("{}{}", value, label));
                    remaining %= unit_secs;
                }
            }
        }

        if parts.is_empty() {
            parts.push("0 seconds".to_string());
        }

        parts.join(", ")
    }

    pub fn get_struct() -> UptimeInfo {
        let boot_time: DateTime<Utc> =
            DateTime::<Utc>::from_timestamp(System::boot_time() as i64, 0)
                .expect("Invalid or out-of-range timestamp");
        let relative_time = StdDuration::from_secs((Utc::now() - boot_time).num_seconds() as u64);

        UptimeInfo {
            boot: DateInfo {
                zone: String::from("UTC"),
                date: boot_time.format("%Y-%m-%d").to_string(),
                time: boot_time.format("%H:%M:%S").to_string(),
            },
            now: DateInfo {
                zone: String::from("UTC"),
                date: Utc::now().format("%Y-%m-%d").to_string(),
                time: Utc::now().format("%H:%M:%S").to_string(),
            },
            relative: format_duration(relative_time).to_string(),
        }
    }

    pub fn get_json() -> Value {
        json!(get_struct())
    }
}

#[derive(Debug, Serialize)]
pub struct TelemetryInfo {
    telemetry: AllInfo,
}

#[derive(Debug, Serialize)]
pub struct AllInfo {
    gpu: gpu::GPUInfo,
    cpu: cpu::CPUInfo,
    memory: memory::MemoryInfo,
    os: os::OSInfo,
    storage: Vec<storage::StorageInfo>,
    network: network::NetworkInfo,
    external: external::ExternalInfo,
    devices: devices::DevicesInfo,
    meta: meta::MetaInfo,
    uptime: uptime::UptimeInfo,
}

pub fn get_struct() -> TelemetryInfo {
    TelemetryInfo {
        telemetry: AllInfo {
            gpu: gpu::get_struct(),
            cpu: cpu::get_struct(),
            memory: memory::get_struct(),
            os: os::get_struct(),
            storage: storage::get_list(),
            network: network::get_struct(),
            external: external::get_struct(),
            devices: devices::get_struct(),
            meta: meta::get_struct(),
            uptime: uptime::get_struct(),
        },
    }
}

pub fn get_json() -> Value {
    json!(get_struct())
}
